import time
import asyncio
from collections import deque

import discord
import aiohttp
from aiohttp import web
from discord.ext import tasks

from . import utils
#import utils

class Client:
    """ The client class used to interact with the Discords.com API """

    def __init__(self, bot: discord.Client, token: str,
                 post: utils.PostObject = None):
        """ Sets up and returns a discordspy Client object

                -- ARGUMENTS --

            :bot: is a discord Client

            :token: is your Discords.com api token

            :post: is a PostObject used to determine how the client will handle
            server count posting
        """

        self.bot = bot
        self.token = token

        self.endpoint = utils.Ratelimit(5, 10)
        self.queue = deque(maxlen=1)
        self.has_webhook = False

        if not post:
            return

        if post.auto:

            async def on_server_change(server):
                await self.post_servers()

            self.bot.add_listener(on_server_change, "on_guild_join")
            self.bot.add_listener(on_server_change, "on_guild_remove")

        elif post.intervals:

            loop = tasks.loop(seconds=post.interval)
            loop(self.post_servers).start()

    def servers(self):
        return len(self.bot.guilds)
    guilds = servers

    async def post_servers(self):
        """ A method to post the server count to the Discords.com API """

        if not self.bot.is_ready():
            return

        timestamp = time.time()
        self.queue.append(timestamp)

        if self.endpoint.is_ratelimited():
            await asyncio.sleep(self.endpoint.until_reset())

        if timestamp not in self.queue:
            return

        url = f"https://discords.com/bots/api/bot/{self.bot.user.id}"
        headers = {"Authorization": self.token}
        data = {"server_count": self.servers()}

        self.endpoint.emulate()

        session = aiohttp.ClientSession()
        async with session.post(url, headers=headers, data=data) as res:
            self.bot.dispatch("discords_server_post", res.status)
            self.bot.dispatch("discords_guild_post", res.status)
        await session.close()

    post_guilds = post_servers

    def webhook(self, port: int = None, auth: str = "", path: str = "/discordswebhook"):
        """ A method to recieve voting webhooks from Discords.com

                -- ARGUMENTS --

            :port: is the port that will be used to listen to webhook requests

            :auth: is the optional auth used to verify that the webhook is
            coming from Discords.com
        """

        if self.has_webhook:
            return

        async def on_post_request(request):

            print("Recieved a request")

            headers = request.headers
            if headers.get("Authorization") != auth:
                return web.Response(status=401)

            try: data = await request.json()
            except: return web.Response(status=400)

            if data.get("type") == "vote":
                self.bot.dispatch("discords_vote", data)
            elif data.get("type") == "test":
                self.bot.dispatch("discords_test", data)
            return web.Response(status=200)

        async def start():

            app = web.Application(loop=self.bot.loop)
            app.router.add_post(path, on_post_request)

            runner = web.AppRunner(app)
            await runner.setup()

            server = web.TCPSite(runner, "0.0.0.0", port)
            await server.start()

        self.bot.loop.create_task(start())
        self.has_webhook = True


# -*- coding: utf-8 -*-

"""
Discords.com API Wrapper
~~~~~~~~~~~~~~~~~~~

A basic wrapper for the Discords.com API.

:copyright: (c) 2021 Judev1
:license: MIT, see LICENSE for more details.
"""

__title__ = "discordspy"
__author__ = "Judev1"
__license__ = "MIT"
__copyright__ = "Copyright 2021 Judev1"
__version__ = "0.1.1"

from .client import Client
from .utils import Post



import time

class PostObject:
    """ A custom class to organise different posting types """

    def __init__(self, post_Type, interval=None):
        """ Sets up and returns a PostObject object

                -- ARGUMENTS --

            :post_Type: is the the number of requests that can be made to an
            endpoint during a period

            :period: is the duration of time in seconds in which the requests
            can be made
        """

        self.auto = False
        self.intervals = False

        if post_Type == "auto":
            self.auto = True
        elif post_Type == "intervals":
            self.intervals = True
            if not interval:
                interval = 1800
            self.interval = interval

class Post:
    """ A non callable class intended to simplify the PostObject class """

    def auto():
        """ A method to create an 'auto' PostObject instance """
        return PostObject("auto")

    def manual():
        """  A method to create a 'manual' PostObject instance """
        return PostObject("manual")

    def intervals(seconds: int = 0, minutes: int = 0, hours: int = 0):
        """  A method to create an 'intervals' PostObject instance. The
            arguments provided are cumulative and so are added together

                -- ARGUMENTS --

            :seconds: is part of the duration of the interval (in seconds)

            :minutes: is part of the duration of the interval (in minutes)

            :hours: is part of the duration of the interval (in hours)
            """
        interval = seconds + minutes*60 + hours*3600
        return PostObject("intervals", interval=interval)

class Ratelimit:
    """ A custom class to emulate api ratelimits """

    period_start = 0
    requests_made = 0

    def __init__(self, requests: float, period: float):
        """ Sets up the ratelimits and returns a Ratelimit object

                -- ARGUMENTS --

            :requests: is the number of requests that can be made to an
            endpoint during a period

            :period: is the duration of time in seconds in which the number of
            requests can be made
        """
        self.requests = requests
        self.period = period

    def emulate(self):
        """ Emulates a request """

        timestamp = time.time()
        if timestamp >= self.period_start + self.period:
            self.period_start = timestamp
            self.requests_made = 1
        elif self.requests_made < self.requests:
            self.requests_made += 1

    def is_ratelimited(self, emulate=False):
        """ Checks if the ratelimit has been reached

                -- ARGUMENTS --

            :emulate: will additionally call the emulate function
        """

        if emulate:
            self.emulate()

        timestamp = time.time()
        if timestamp >= self.period_start + self.period:
            return False
        elif self.requests_made < self.requests:
            return False
        return True

    def until_reset(self):
        """ Returns the amount of time until the endpoint is no longer
            ratelimited
        """

        timestamp = time.time()

        seconds = (self.period_start + self.period) - timestamp
        if timestamp >= self.period_start + self.period:
            seconds = 0

        return seconds


from setuptools import setup

with open("discordspy/README.md", "r") as file: description = file.read()

setup(
    name="discordspy",
    version="0.1.1",
    description="A basic API wrapper for Discords.com",
    long_description=description,
    long_description_content_type="text/markdown",
    url="https://github.com/judev1/discords",
    author="Jude BC",
    author_email="jude.version1.0@gmail.com",
    license="MIT",
    packages=["discordspy"],

    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
)
